/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005-2008. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("net.jazz.ajax.ui.Workbench");

dojo.require("dijit.Dialog");

dojo.require("net.jazz.ajax.core.internal.loader");
dojo.require("net.jazz.ajax.core.internal.Uri");
dojo.require("net.jazz.ajax.core.internal.UriMonitor");
dojo.require("net.jazz.ajax.core.internal.util");
dojo.require("net.jazz.ajax.toolbox.lang.internal.Lang");
dojo.require("net.jazz.ajax.ui.ActionRegistry");
dojo.require("net.jazz.ajax.ui.ResourceRegistry");
dojo.require("net.jazz.ajax.ui.internal.Dialog");
dojo.require("net.jazz.ajax.ui.internal.PageRegistry");

dojo.require("dojo.i18n"); 
dojo.requireLocalization("net.jazz.ajax.ui.internal", "JAFMessages");

var messages;

dojo.declare("net.jazz.ajax.ui.Workbench", null, {
	
	// summary:
	//		provides top-level UI methods to Jazz Ajax Framework applications
	// usage notes:
	//		clients gain access to the running workbench via
	//		net.jazz.ajax.PlatformUI.getWorkbench(); do not directly instantiate
	//		nor subclass
	
	getActionRegistry: function() {
		// summary:
		//		returns the action registry for the workbench
		if (!this._actionRegistry) {
			this._actionRegistry = new ActionRegistry();
		}
		return this._actionRegistry;
	},
	
	getResourceRegistry: function() {
		// summary:
		//		returns the resource registry for the workbench
		if (!this._ResourceRegistry) {
			this._ResourceRegistry = new ResourceRegistry();
		}
		return this._ResourceRegistry;
	},

	hideMessageArea: function() {
		// summary:
		//		hides the message area; usually called at the end of a long-
		//		running operation
		dojo.publish(net.jazz.ajax.internal.events.hideMessageArea, []);
	},
	
	orderPages: function(/*Function*/sortFunc) {
		// summary:
		//		Internally orders pages based on the logic of the function
		//		sortFunc, which follows the same rules as Array#sort.
		// usage note:
		//		Calling this method causes any Jazz Ajax Framework widgets that
		//		use pages to reorder themselves (e.g. PageList).
		PageRegistry.setSortFunction(sortFunc);
		dojo.publish(net.jazz.ajax.internal.events.orderPages, []);
	},
	rootNode: function() {
		// summary:
		//		returns the root node of the workbench
		return dojo.byId("net-jazz-ajax-WorkbenchRoot");
	},
	registerBodyClickHandler: function(func, priority) {
		var handlers = this._bodyClickHandlers;
		var newHandler = {f: func, p: priority};
		this._arrayInsertByPriority(handlers,newHandler);
		return newHandler;
	},
	registerOnBeforeUnloadHandler: function(func, priority) {
		var handlers = this._onBeforeUnloadHandlers;
		var newHandler = {f: func, p: priority};
		this._arrayInsertByPriority(handlers,newHandler);
		return newHandler;
	},
	unregisterBodyClickHandler: function(handler) {
		return this._arrayRemoveByReference(this._bodyClickHandlers, handler);
	},
	unregisterOnBeforeUnloadHandler: function(handler) {
		return this._arrayRemoveByReference(this._onBeforeUnloadHandlers, handler);		
	},	
	viewDefaultPage: function() {
		// summary:
		//		displays the default page for the application, currently just
		//		the first page in the page list
		var pages = PageRegistry.getPages();
		if (pages.length === 0) {
			console.warn("Workbench#viewDefaultPage: Cannot view default page: no pages found.");
		} else {
			this._viewPageActionImpl({id: pages[0].id});
		}
	},
	
	viewPage: function(pageId, cb) {
		// summary:
		//		displays the page with the id specified by 'pageId'
		// usage note:
		//      This api potentially loads a page widget asynchronously and thus
		//      must be passed a callback function if there is functionality that
		//      has to occur after the page has loaded. 
		try {
			checkNotNull(pageId, "pageId");	
		} catch (e) {
			this.currentPageId = null;
			dojo.publish(net.jazz.ajax.internal.events.viewErrorPage, [messages.noPageId, e]);
			return;
		}
		
		var page = PageRegistry.findPageWithId(pageId);
		
		if (page == null) {
			this.currentPageId = null;
			dojo.publish(net.jazz.ajax.internal.events.viewErrorPage, [messages.pageNotFound, messages.pageId + pageId]);
			return;
		}
		
		// try to get page widget from the cache
		var pageWidget = this._pageWidgetCache[page.id];
		
		// if the page widget wasn't in the cache, lazily load the page from the
		// server and store it in the page cache
		if (!pageWidget) {
			this.showMessage(dojo.string.substitute(messages.loading, [page.name]));
			var _self = this;
			load_async(page.widget, function() {
				var pageConstructor = dojo.getObject(page.widget);
				pageWidget = new pageConstructor();
				_self._pageWidgetCache[page.id] = pageWidget;
				_self.hideMessageArea();
				page.domNode = pageWidget.domNode;
				_self.currentPageId = page.id;
				dojo.publish(net.jazz.ajax.internal.events.viewPage, [page]);
				if (cb != undefined && dojo.isFunction(cb)) {
					cb();
				}
			});
		}
		else {
			page.domNode = pageWidget.domNode;
			this.currentPageId = page.id;
			// publish view page topic
			dojo.publish(net.jazz.ajax.internal.events.viewPage, [page]);
			if (cb != undefined && dojo.isFunction(cb)) {
				cb();
			}
		}
	},
	
	setTitle: function(titleStr) {
		// summary:
		//		sets the page title; usually called at the beginning of an
		//		action invocation 
		checkNotNull(titleStr, "titleStr");
		titleStr = dojo.trim(titleStr);
		document.title = titleStr ?
			titleStr + SEPARATOR + this._defaultTitle :
			this._defaultTitle;
		this._titleWasSet = true;
	},
	
	showMessage: function(message) {
		// summary:
		//		displays the message specified by 'message'; usually called at
		//		the beginning of long-running operations
		checkNotNull(message, "message");
		dojo.publish(net.jazz.ajax.internal.events.showMessage, [message]);
	},
	
	//
	// Internal properties and methods; do not call directly
	//
	
	constructor: function() {
		messages = dojo.i18n.getLocalization("net.jazz.ajax.ui.internal", "JAFMessages");
		
		this._pageWidgetCache = {};
		this.getActionRegistry().registerAction("jazz.viewPage", this, "_viewPageActionImpl");
		
		// initialize body click handlers
		this._bodyClickHandlers = new Array();
		this._onBeforeUnloadHandlers = new Array();
		this.registerBodyClickHandler(dojo.hitch(this, this._notifyURIMonitorHandler), /*priority*/ 1); //lowest priority
		this.registerBodyClickHandler(dojo.hitch(this, this._hasResourceHandler), /*priority*/ 2);
		// intercept body click
		dojo.connect(dojo.body(), "onclick", this, "_handleBodyClick");
		window.onbeforeunload = dojo.hitch(this,this._handleOnBeforeUnload);
	},
	_arrayInsertByPriority: function (array, newEntry) {
		//newEntry: {p: priority, func: callback function}
		for (var i=0; i<array.length; i++) {
			if (newEntry.p > array[i].p) {
				array.splice(i, 0, newEntry);
				return;
			}
		}
		// if we didn't insert the new handler, just append it to the list
		array.push(newEntry);
	},
	_arrayRemoveByReference: function (array, entry) {
		for (var i = 0; i < array.length;i++) {
			if (entry == array[i]) {
				array.splice(i,1);
				return true;	
			}
		}
		return false;
	},	
	// the default default (sic) title if none other is specified
	_defaultTitle: "Jazz",
	
	_handleBodyClick: function(e) {
		if (e.altKey || e.ctrlKey || e.shiftKey) {
			return;
		}
		if (e.target) {
			if (e.target.tagName === "A") {
				var href = e.target.href;
			} else if (e.target.parentNode && e.target.parentNode.tagName === "A") {
				var href = e.target.parentNode.href;
			} else if (e.target.parentNode && e.target.parentNode.parentNode && e.target.parentNode.parentNode.tagName === "A") {
				var href = e.target.parentNode.parentNode.href;
			} else {
				return;
			}
		} else {
			return;
		}
		var handlers = this._bodyClickHandlers;
		var urlObj = new dojo._Url(href);
		for (var i = 0; i < handlers.length; i++) {
			var shouldVeto = handlers[i].f(urlObj);
			if (shouldVeto) {
				dojo.stopEvent(e);
				break;
			}
		}
	},
	
	// for occasions when body click logic must be fired programmatically.
	// returns true if event should be vetoed.
	shouldVetoNavigation: function(href) {
		var urlObj = new dojo._Url(href);
		var winLoc = new dojo._Url(window.location.href);
		if (urlObj.fragment && ((!urlObj.host && !urlObj.path) || (urlObj.path && urlObj.path === winLoc.path))) { // we're navigating within web UI.
			var handlers = this._bodyClickHandlers;
			for (var i = 0; i < handlers.length; i++)
				if (handlers[i].f(urlObj)) return true;
			return false;
		} else {
			var message = this._getOnBeforeUnloadMessage();
			if (message !== false) {
				//fake the onBeforeUnload dialog
				if (confirm(messages.onBeforeUnloadPrefix + message + messages.onBeforeUnloadSuffix)) {
					this._onBeforeUnloadHandlers = new Array(); //The user has agreed to navigate away, so short circuit the real onBeforeUnload
					return false;
				} else {
					return true;
				}
			} else 
				return false;
		}
	},
	_handleOnBeforeUnload: function (e) {
		if (!e) e = window.event;  // need to fetch the event object explicitly in IE
		var message = this._getOnBeforeUnloadMessage();
		if (message !== false) {
			e.returnValue = message;
		}
	},
	_getOnBeforeUnloadMessage: function() {
		var handlers = this._onBeforeUnloadHandlers;
		var msgs = new Array();
		for (var i = 0; i < handlers.length; i++) {
			var message = handlers[i].f(/*isOnBeforeUnload*/ true); 
			if (message) msgs.push(message);
		}
		if (msgs.length > 0) { // if any messages were returned, format the output and show the prompt
			msgs.push(messages.changesWillBeDiscarded);
			var output = "";
			for (var i = 0; i < msgs.length; i++) {
				if (msgs.length > 2 && i != msgs.length - 1) output += "- ";
				output += msgs[i];
				if (i == msgs.length - 2) {
					if (msgs.length == 2) {
						output += " ";
					} else {
						output += "\n\n";
					} 
				} else if (i != msgs.length - 1) {
					output += "\n";
				}
			}
			return output;
		}
		return false;
	},
	_notifyURIMonitorHandler: function(uriObj) {
		var winLoc = new dojo._Url(window.location.href);
		
		if(!((!uriObj.scheme || uriObj.scheme === winLoc.scheme) && (!uriObj.host || uriObj.host === winLoc.host))) {
		   	return false; // different domain, so do nothing
		}
		if (uriObj.path != winLoc.path) {
			return false; // ignore different paths
		}
		
		if (isActionUri(uriObj)) {
			window.location.href = "#" + uriObj.fragment;
			this._uriMonitor.pollLocation();
		}
		return false;
	},
	
	_hasResourceHandler: function(uriObj) {
		
		var winLoc = new dojo._Url(window.location.href);
		
		if(!(uriObj.scheme === winLoc.scheme && uriObj.host === uriObj.host)) {
		   	return false; // different domain, so do nothing
		}
		if (uriObj.path === winLoc.path) {
			return false; // same exact base URI, so do nothing
		}
		
		var urlSuffix = trimPrefix(uriObj);
		var hasHandler = this.getResourceRegistry().findPatternForUrl(urlSuffix);
		if (!hasHandler) {
			return false; // Should not veto, early return
		}
		
		window.location.href = "#" + urlSuffix;
		return true; // Redirecting to resource-centric URI so should veto body click
	},
	
	_processUri: function(uri) {
		checkNotNull(uri, "uri");
		
		// reset state so that we can check later if the code we called set the
		// title or not
		this._titleWasSet = false;
		
		var handled = false;
		
		// action-based URIs have the form:
		// #action=[action-id]&arg1=val1
		if (isActionUri(uri)) {
			this._processActionUri(uri);
			handled = true;
		}
		
		// action-based URIs have the form:
		// #action=[action-id]&arg1=val1
		if (!handled && uri.fragment){
			var patternMatch = this.getResourceRegistry().findPatternForUrl(uri.fragment);
			if (patternMatch) {
				this._processResourceUri(uri, patternMatch);
				handled = true;
			} else if (patternMatch === false) {
				this.currentPageId = null;
				//not an action or resource based URL, but has a hash value
				dojo.publish(net.jazz.ajax.internal.events.viewErrorPage, [messages.invalidUrl]);
				handled = true;
			}
		}
		
		// if neither an action nor a resourceHandler handled the URI, bail out
		// and show the default page
		if (!handled) {
			this.viewDefaultPage();
		}
		
		// if the code we called didn't set the title, just show the default
		if (!this._titleWasSet) {
			this.setTitle("");
		}
	},
	
	_processActionUri: function(uri) {
		checkNotNull(uri, "uri");
		
		var fragmentArgs = getUriFragmentHash(uri);
		
		var actionId = fragmentArgs.action;
		
		// _processActionUri should not have been called if the URI did not
		// contain an action param
		checkNotNull(actionId, "actionId");
		
		var actionEntry = this.getActionRegistry()._getActionEntry(actionId);
		
		if (!actionEntry) {
			this.currentPageId = null;
			dojo.publish(net.jazz.ajax.internal.events.viewErrorPage, [messages.invalidAction, messages.action + actionId]);
			return;
		}
		
		// if this action is supposed to run in the context of a particular page
		// then show the page.
		if (actionEntry._requiresPage) {
			var pageId = actionEntry._requiresPage;
			var self = this;
			this.viewPage(pageId, function() {
				self._runAction("action", actionId, fragmentArgs);
			});
		}
		else {
			this._runAction("action", actionId, fragmentArgs);
		}
		
	},
	
	_processResourceUri: function(/*Object*/ uri, /*String*/ patternMatch) {
		var fragment = decodeURIComponent(uri.fragment);
		
		var argObj = this._parseResourceUri(fragment);
		
		var resourceHandlerModuleName = this.getResourceRegistry().findModuleNameForPattern(patternMatch);
		
		if (resourceHandlerModuleName) {
			var self = this;
			this.viewPage(resourceHandlerModuleName, function() {
				self = self._runAction("resourceHandler", patternMatch, argObj);
			});
		} else {
			self = self._runAction("resourceHandler", patternMatch, argObj);
		}
	},
	
	setResourceUrlParser: function(parseFunction) {
		if (!dojo.isFunction(parseFunction)) {
			throw new Error("Workbench#setResourceUrlParser: parameter 'parseFunction' not a function (was: " + parseFunction + " )");
		}
		// replace default parser with custom parser
		this._parseResourceUri = parseFunction;
	},
	
	_parseResourceUri: function(url) {
		// default implementation
		return {url: url};
	},
	_runAction: function(type, id, objArg) {
		if (type === "action") {
			var action = this.getActionRegistry().findAction(id);
		} else if (type === "resourceHandler") {
			var actionArr = this.getResourceRegistry().findHandlersForPattern(id);
			if (actionArr.length === 0) {
				throw new Error("No implementation bound to pattern " + id);
			}
			if (actionArr.length > 1) {
				console.warn("More than one action bound to pattern: " + id + "; using first one (multiple handlers not yet supported).")
			}
			action = actionArr[0];
		} else {
			throw new Error("Invalid type: " + type);
		}
		callContributedCode( function() {
			action.run(objArg);
		});
	},
	
	_start: function() {
		// guard against multiple calls
		if(this.__isStarted) {
			return;
		}
		this.__isStarted = true;
		
		// summary:
		//		informs the workbench to begin processing events
		var self = this;
		// create URI monitor and process first URI
		this._uriMonitor = createUriMonitor(function(uri) {
			self._processUri(uri);
		});
		this._processUri(new dojo._Url(location.href));
		dojo.body().removeChild(dojo.byId("net-jazz-ajax-InitialLoadMessage"));
	},
	
	_setDefaultTitle: function(titleStr) {
		// summary:
		//		sets the default page title that is always prepended before the
		//		argument sent to Workbench#setTitle 
		checkNotEmpty(titleStr, "titleStr");
		this._defaultTitle = titleStr;
	},
	
	_viewPageActionImpl: function(/*Object*/kwArgs) {
		// description:
		//		Implementation for "jazz.viewPage" action. Loads the page widget
		//		(if necessary), displays it in the web workbench, and if this is
		//		the first time the user has visited the page, runs the default
		//		action.
		// kwArgs:
		//		action args; must contain an 'id' property whose value
		//		corresponds to the id of an installed page extension.
		try {
			checkNotNull(kwArgs.id, "kwArgs.id");
		} catch (e) {
			this.currentPageId = null;
			dojo.publish(net.jazz.ajax.internal.events.viewErrorPage, [messages.noPageId, e]);
			return;
		}
		// show the page
		var _self = this;
		this.viewPage(kwArgs.id, function() {
			var page = PageRegistry.findPageWithId(kwArgs.id);
			
			// set the document's title to the page's name
			_self.setTitle(page.name);
			
			// run the page's default action
			_self._runAction("action", page.defaultAction, {});
		});
	}
});

var ActionRegistry = net.jazz.ajax.ui.ActionRegistry;
var PageRegistry = net.jazz.ajax.ui.internal.PageRegistry;
var ResourceRegistry = net.jazz.ajax.ui.ResourceRegistry;

var callContributedCode = net.jazz.ajax.core.internal.util.callContributedCode;
var checkNotEmpty = net.jazz.ajax.toolbox.lang.checkNotEmpty;
var checkNotNull = net.jazz.ajax.toolbox.lang.checkNotNull;
var createUriMonitor = net.jazz.ajax.core.internal.createUriMonitor;
var getUriFragmentHash = net.jazz.ajax.core.internal.getUriFragmentHash;
var load_async = net.jazz.ajax.core.internal.loader.load_async;

var NO_INIT_METHOD_MSG = "Cannot start application (no 'init' method found).";
var SEPARATOR = " - ";

function isActionUri(/*Object*/uri) {
	// summary:
	//		returns true if the specified URI is an action-style URI (i.e. it
	//		contains an action=[action-id] part), otherwise false.
	// uri:
	//		a Dojo URI object 
	checkNotNull(uri, "uri");
	
	if (!uri.fragment) return false;
	
	return uri.fragment.indexOf("action=") === 0;
}

function defaultResourceUriParser(uriStr) {
	checkNotNull(uriStr, "uriStr");
	return {uri: uriStr};
}

function trimPrefix(uriObj) {
	var contextRoot = net.jazz.ajax._contextRoot || "";
	
	var path = uriObj.path;
	if (path.indexOf(contextRoot) === 0) {
		path = path.substring(contextRoot.length);
	}
	
	var queryStr = uriObj.query ? encodeURIComponent("?" + uriObj.query) : "";
	
	return path + queryStr;
}

})();
